/*
(C) 2002 Advanced Micro Devices, Inc. 
** YOUR USE OF THIS LIBRARY IS SUBJECT TO THE TERMS 
    AND CONDITIONS OF THE GNU LESSER GENERAL PUBLIC 
    LICENSE FOUND IN THE "README" FILE THAT IS INCLUDED WITH 
    THIS LIBRARY**
*/


#include "libm_amd.h"
#include "libm_util_amd.h"

#define USE_NAN_WITH_FLAGS
#define USE_VAL_WITH_FLAGS
#include "libm_inlines_amd.h"
#undef USE_NAN_WITH_FLAGS
#undef USE_VAL_WITH_FLAGS

/* sin(x) approximation valid on the interval [-pi/4,pi/4]. */
static inline double sin_piby4(double x, double xx)
{
  /* Taylor series for sin(x) is x - x^3/3! + x^5/5! - x^7/7! ...
                          = x * (1 - x^2/3! + x^4/5! - x^6/7! ...
                          = x * f(w)
     where w = x*x and f(w) = (1 - w/3! + w^2/5! - w^3/7! ...
     We use a minimax approximation of (f(w) - 1) / w
     because this produces an expansion in even powers of x.
     If xx (the tail of x) is non-zero, we add a correction
     term g(x,xx) = (1-x*x/2)*xx to the result, where g(x,xx)
     is an approximation to cos(x)*sin(xx) valid because
     xx is tiny relative to x.
  */
  static const double
    c1 = -0.166666666666666646259241729,
    c2 = 0.833333333333095043065222816e-2,
    c3 = -0.19841269836761125688538679e-3,
    c4 = 0.275573161037288022676895908448e-5,
    c5 = -0.25051132068021699772257377197e-7,
    c6 = 0.159181443044859136852668200e-9;
  double x2, x3, r;
  x2 = x * x;
  x3 = x2 * x;
  r = (c2 + x2 * (c3 + x2 * (c4 + x2 * (c5 + x2 * c6))));
  if (xx == 0.0)
    return x + x3 * (c1 + x2 * r);
  else
    return x - ((x2 * (0.5 * xx - x3 * r) - xx) - x3 * c1);
}

/* cos(x) approximation valid on the interval [-pi/4,pi/4]. */
static inline double cos_piby4(double x, double xx)
{
  /* Taylor series for cos(x) is 1 - x^2/2! + x^4/4! - x^6/6! ...
                          = f(w)
     where w = x*x and f(w) = (1 - w/2! + w^2/4! - w^3/6! ...
     We use a minimax approximation of (f(w) - 1 + w/2) / (w*w)
     because this produces an expansion in even powers of x.
     If xx (the tail of x) is non-zero, we subtract a correction
     term g(x,xx) = x*xx to the result, where g(x,xx)
     is an approximation to sin(x)*sin(xx) valid because
     xx is tiny relative to x.
  */
  double r, x2, t;
  static const double
    c1 = 0.41666666666666665390037e-1,
    c2 = -0.13888888888887398280412e-2,
    c3 = 0.248015872987670414957399e-4,
    c4 = -0.275573172723441909470836e-6,
    c5 = 0.208761463822329611076335e-8,
    c6 = -0.113826398067944859590880e-10;

  x2 = x * x;
  r = 0.5 * x2;
  t = 1.0 - r;
  return t + ((((1.0 - t) - r) - x * xx) + x2 * x2 *
              (c1 + x2 * (c2 + x2 * (c3 + x2 * (c4 + x2 * (c5 + x2 * c6))))));
}

void __sincos(double x, double *s, double *c)
{
  double r, rr;
  int region, xneg;

  unsigned long ux, ax;
  GET_BITS_DP64(x, ux);
  ax = (ux & ~SIGNBIT_DP64);
  if (ax <= 0x3fe921fb54442d18) /* abs(x) <= pi/4 */
    {
      if (ax < 0x3f20000000000000) /* abs(x) < 2.0^(-13) */
        {
          if (ax < 0x3e40000000000000) /* abs(x) < 2.0^(-27) */
	    {
	      if (ax == 0x0000000000000000)
                {
                  *s = x;
                  *c = 1.0;
                }
              else
                {
                  *s = x;
                  *c = val_with_flags(1.0, AMD_F_INEXACT);
                }
	    }
          else
            {
              *s = x - x*x*x*0.166666666666666666;
              *c = 1.0 - x*x*0.5;
            }
        }
      else
        {
          *s = sin_piby4(x, 0.0);
          *c = cos_piby4(x, 0.0);
        }
      return;
    }
  else if ((ux & EXPBITS_DP64) == EXPBITS_DP64)
    {
      /* x is either NaN or infinity */
      if (ux & MANTBITS_DP64)
        /* x is NaN */
        *s = *c = x + x; /* Raise invalid if it is a signalling NaN */
      else
        /* x is infinity. Return a NaN */
        *s = *c = nan_with_flags(AMD_F_INVALID);
      return;
    }

  xneg = (ax != ux);


  if (xneg)
    x = -x;

  /* Reduce x into range [-pi/4,pi/4] */
  __remainder_piby2(x, &r, &rr, &region);

  if (xneg)
    {
      switch (region)
        {
        default:
        case 0:
          *s = -sin_piby4(r, rr);
          *c = cos_piby4(r, rr);
          break;
        case 1:
          *s = -cos_piby4(r, rr);
          *c = -sin_piby4(r, rr);
          break;
        case 2:
          *s = sin_piby4(r, rr);
          *c = -cos_piby4(r, rr);
          break;
        case 3:
          *s = cos_piby4(r, rr);
          *c = sin_piby4(r, rr);
          break;
        }
    }
  else
    {
      switch (region)
        {
        default:
        case 0:
          *s = sin_piby4(r, rr);
          *c = cos_piby4(r, rr);
          break;
        case 1:
          *s = cos_piby4(r, rr);
          *c = -sin_piby4(r, rr);
          break;
        case 2:
          *s = -sin_piby4(r, rr);
          *c = -cos_piby4(r, rr);
          break;
        case 3:
          *s = -cos_piby4(r, rr);
          *c = sin_piby4(r, rr);
          break;
        }
    }
  return;
}

double __sin(double x)
{
  double r, rr;
  int region, xneg;

  unsigned long ux, ax;
  GET_BITS_DP64(x, ux);
  ax = (ux & ~SIGNBIT_DP64);
  if (ax <= 0x3fe921fb54442d18) /* abs(x) <= pi/4 */
    {
      if (ax < 0x3f20000000000000) /* abs(x) < 2.0^(-13) */
        {
          if (ax < 0x3e40000000000000) /* abs(x) < 2.0^(-27) */
	    {
	      if (ax == 0x0000000000000000)
                return x;
              else
                return val_with_flags(x, AMD_F_INEXACT);
	    }
          else
            return x - x*x*x*0.166666666666666666;
        }
      else
        return sin_piby4(x, 0.0);
    }
  else if ((ux & EXPBITS_DP64) == EXPBITS_DP64)
    {
      /* x is either NaN or infinity */
      if (ux & MANTBITS_DP64)
        /* x is NaN */
        return x + x; /* Raise invalid if it is a signalling NaN */
      else
        /* x is infinity. Return a NaN */
        return nan_with_flags(AMD_F_INVALID);
    }
  xneg = (ax != ux);


  if (xneg)
    x = -x;

  /* Reduce x into range [-pi/4,pi/4] */
  __remainder_piby2(x, &r, &rr, &region);

  if (xneg)
    {
      switch (region)
        {
        default:
        case 0: return -sin_piby4(r, rr);
        case 1: return -cos_piby4(r, rr);
        case 2: return sin_piby4(r, rr);
        case 3: return cos_piby4(r, rr);
        }
    }
  else
    {
      switch (region)
        {
        default:
        case 0: return sin_piby4(r, rr);
        case 1: return cos_piby4(r, rr);
        case 2: return -sin_piby4(r, rr);
        case 3: return -cos_piby4(r, rr);
        }
    }
}

double __cos(double x)
{
  double r, rr;
  int region, xneg;

  unsigned long ux, ax;
  GET_BITS_DP64(x, ux);
  ax = (ux & ~SIGNBIT_DP64);
  if (ax <= 0x3fe921fb54442d18) /* abs(x) <= pi/4 */
    {
      if (ax < 0x3f20000000000000) /* abs(x) < 2.0^(-13) */
        {
          if (ax < 0x3e40000000000000) /* abs(x) < 2.0^(-27) */
            {
              if (ax == 0x0000000000000000) /* abs(x) = 0.0 */
                return 1.0;
              else
                return val_with_flags(1.0, AMD_F_INEXACT);
            }
          else
            return 1.0 - x*x*0.5;
        }
      else
        return cos_piby4(x, 0.0);
    }
  else if ((ux & EXPBITS_DP64) == EXPBITS_DP64)
    {
      /* x is either NaN or infinity */
      if (ux & MANTBITS_DP64)
        /* x is NaN */
        return x + x; /* Raise invalid if it is a signalling NaN */
      else
        /* x is infinity. Return a NaN */
        return nan_with_flags(AMD_F_INVALID);
    }
  xneg = (ax != ux);


  if (xneg)
    x = -x;

  /* Reduce x into range [-pi/4,pi/4] */
  __remainder_piby2(x, &r, &rr, &region);

  switch (region)
    {
    default:
    case 0: return cos_piby4(r, rr);
    case 1: return -sin_piby4(r, rr);
    case 2: return -cos_piby4(r, rr);
    case 3: return sin_piby4(r, rr);
    }
}

weak_alias (__sin, sin)
weak_alias (__cos, cos)
weak_alias (__sincos, sincos)
