/*
(C) 2002 Advanced Micro Devices, Inc. 
** YOUR USE OF THIS LIBRARY IS SUBJECT TO THE TERMS 
    AND CONDITIONS OF THE GNU LESSER GENERAL PUBLIC 
    LICENSE FOUND IN THE "README" FILE THAT IS INCLUDED WITH 
    THIS LIBRARY**
*/

#include "libm_amd.h"
#include "libm_util_amd.h"

#define USE_SPLITEXP
#define USE_SCALEDOUBLE_2
#define USE_SCALEDOUBLE_3
#define USE_SCALEDOWNDOUBLE
#define USE_INFINITY_WITH_FLAGS
#define USE_ZERO_WITH_FLAGS
#define USE_NAN_WITH_FLAGS
#define USE_VAL_WITH_FLAGS
#include "libm_inlines_amd.h"
#undef USE_SPLITEXP
#undef USE_SCALEDOUBLE_2
#undef USE_SCALEDOUBLE_3
#undef USE_SCALEDOWNDOUBLE
#undef USE_INFINITY_WITH_FLAGS
#undef USE_ZERO_WITH_FLAGS
#undef USE_NAN_WITH_FLAGS
#undef USE_VAL_WITH_FLAGS

/* Deal with errno for out-of-range result */
#include "libm_errno_amd.h"
static inline double retval_errno_erange_overflow(double x, double y, int sign)
{
  struct exception exc;
  exc.arg1 = x;
  exc.arg2 = y;
  exc.type = OVERFLOW;
  exc.name = (char *)"pow";
  if (_LIB_VERSION == _SVID_)
    {
      if (sign == 1)
        exc.retval = HUGE;
      else /* sign = -1 */
        exc.retval = -HUGE;
    }
  else
    {
      if (sign == 1)
        exc.retval = infinity_with_flags(AMD_F_OVERFLOW);
      else /* sign == -1 */
        exc.retval = -infinity_with_flags(AMD_F_OVERFLOW);
    }
  if (_LIB_VERSION == _POSIX_)
    __set_errno(ERANGE);
  else if (!matherr(&exc))
    __set_errno(ERANGE);
  return exc.retval;
}

static inline double retval_errno_erange_underflow(double x, double y, int sign)
{
  struct exception exc;
  exc.arg1 = x;
  exc.arg2 = y;
  exc.type = UNDERFLOW;
  exc.name = (char *)"pow";
  if (sign == 1)
    exc.retval = zero_with_flags(AMD_F_UNDERFLOW | AMD_F_INEXACT);
  else /* sign == -1 */
    exc.retval = -zero_with_flags(AMD_F_UNDERFLOW | AMD_F_INEXACT);
  if (_LIB_VERSION == _POSIX_)
    __set_errno(ERANGE);
  else if (!matherr(&exc))
    __set_errno(ERANGE);
  return exc.retval;
}

/* Deal with errno for out-of-range arguments */
static inline double retval_errno_edom(double x, double y, int type)
{
  struct exception exc;
  exc.arg1 = x;
  exc.arg2 = y;
  exc.type = DOMAIN;
  exc.name = (char *)"pow";
  if (_LIB_VERSION == _SVID_)
    exc.retval = 0.0;
  else if (type == 1)
    exc.retval = infinity_with_flags(AMD_F_DIVBYZERO);
  else if (type == 2)
    exc.retval = -infinity_with_flags(AMD_F_DIVBYZERO);
  else /* type == 3 */
    exc.retval = nan_with_flags(AMD_F_INVALID);
  if (_LIB_VERSION == _POSIX_)
    __set_errno (EDOM);
  if (!matherr(&exc))
    {
      if (_LIB_VERSION == _SVID_)
        (void)fputs("pow: DOMAIN error\n", stderr);
      __set_errno(EDOM);
    }
  return exc.retval;
}

/* Computes the exact product of x and y, the result being the
   nearly doublelength number (z,zz) */
static void mul12(double x, double y,
                  double *z, double *zz)
{
  double hx, tx, hy, ty;
  /* Split x into hx (head) and tx (tail). Do the same for y. */
  unsigned long u;
  GET_BITS_DP64(x, u);
  u &= 0xfffffffff8000000;
  PUT_BITS_DP64(u, hx);
  tx = x - hx;
  GET_BITS_DP64(y, u);
  u &= 0xfffffffff8000000;
  PUT_BITS_DP64(u, hy);
  ty = y - hy;
  *z = x * y;
  *zz = (((hx * hy - *z) + hx * ty) + tx * hy) + tx * ty;
}

/* Computes the doublelength product of (x,xx) and (y,yy), the result
   being the doublelength number (z,zz) */
static void mul2(double x, double xx, double y, double yy,
                 double *z, double *zz)
{
  double c, cc;
  mul12(x, y, &c, &cc);
  cc = x * yy + xx * y + cc;
  *z = c + cc;
  *zz = (c - *z) + cc;
}

double __pow(double x, double y)
{

  /* Arrays log2_lead_table and log2_tail_table contain
     leading and trailing parts respectively of precomputed
     values of natural log2(1+i/64), for i = 0, 1, ..., 64.
     ln_lead_table contains the first 24 bits of precision,
     and ln_tail_table contains a further 53 bits precision. */

  static const double log2_lead_table[65] = {
    0.00000000000000000000E+00,  /* 0x0000000000000000 */
    2.23678126931190490723E-02,  /* 0x3F96E79680000000 */
    4.43941168487071990967E-02,  /* 0x3FA6BAD360000000 */
    6.60891830921173095703E-02,  /* 0x3FB0EB3880000000 */
    8.74628350138664245605E-02,  /* 0x3FB663F6E0000000 */
    1.08524456620216369629E-01,  /* 0x3FBBC84240000000 */
    1.29283010959625244141E-01,  /* 0x3FC08C5880000000 */
    1.49747118353843688965E-01,  /* 0x3FC32AE9E0000000 */
    1.69924989342689514160E-01,  /* 0x3FC5C01A20000000 */
    1.89824551343917846680E-01,  /* 0x3FC84C2BC0000000 */
    2.09453359246253967285E-01,  /* 0x3FCACF5E20000000 */
    2.28818684816360473633E-01,  /* 0x3FCD49EE40000000 */
    2.47927501797676086426E-01,  /* 0x3FCFBC16A0000000 */
    2.66786515712738037109E-01,  /* 0x3FD11307C0000000 */
    2.85402208566665649414E-01,  /* 0x3FD24407A0000000 */
    3.03780734539031982422E-01,  /* 0x3FD37124C0000000 */
    3.21928083896636962891E-01,  /* 0x3FD49A7840000000 */
    3.39849978685379028320E-01,  /* 0x3FD5C01A20000000 */
    3.57551991939544677734E-01,  /* 0x3FD6E221C0000000 */
    3.75039428472518920898E-01,  /* 0x3FD800A560000000 */
    3.92317414283752441406E-01,  /* 0x3FD91BBA80000000 */
    4.09390926361083984375E-01,  /* 0x3FDA337600000000 */
    4.26264733076095581055E-01,  /* 0x3FDB47EBE0000000 */
    4.42943483591079711914E-01,  /* 0x3FDC592FA0000000 */
    4.59431618452072143555E-01,  /* 0x3FDD6753E0000000 */
    4.75733429193496704102E-01,  /* 0x3FDE726AA0000000 */
    4.91853088140487670898E-01,  /* 0x3FDF7A8560000000 */
    5.07794618606567382812E-01,  /* 0x3FE03FDA80000000 */
    5.23561954498291015625E-01,  /* 0x3FE0C10500000000 */
    5.39158761501312255859E-01,  /* 0x3FE140C9E0000000 */
    5.54588794708251953125E-01,  /* 0x3FE1BF3100000000 */
    5.69855570793151855469E-01,  /* 0x3FE23C41C0000000 */
    5.84962487220764160156E-01,  /* 0x3FE2B80340000000 */
    5.99912822246551513672E-01,  /* 0x3FE3327C60000000 */
    6.14709794521331787109E-01,  /* 0x3FE3ABB3E0000000 */
    6.29356563091278076172E-01,  /* 0x3FE423B060000000 */
    6.43856167793273925781E-01,  /* 0x3FE49A7840000000 */
    6.58211469650268554688E-01,  /* 0x3FE5101180000000 */
    6.72425329685211181641E-01,  /* 0x3FE5848220000000 */
    6.86500489711761474609E-01,  /* 0x3FE5F7CFE0000000 */
    7.00439691543579101562E-01,  /* 0x3FE66A0080000000 */
    7.14245498180389404297E-01,  /* 0x3FE6DB1960000000 */
    7.27920413017272949219E-01,  /* 0x3FE74B1FC0000000 */
    7.41466939449310302734E-01,  /* 0x3FE7BA18E0000000 */
    7.54887461662292480469E-01,  /* 0x3FE82809C0000000 */
    7.68184304237365722656E-01,  /* 0x3FE894F740000000 */
    7.81359672546386718750E-01,  /* 0x3FE900E600000000 */
    7.94415831565856933594E-01,  /* 0x3FE96BDAC0000000 */
    8.07354867458343505859E-01,  /* 0x3FE9D5D9E0000000 */
    8.20178925991058349609E-01,  /* 0x3FEA3EE7E0000000 */
    8.32889974117279052734E-01,  /* 0x3FEAA708E0000000 */
    8.45490038394927978516E-01,  /* 0x3FEB0E4120000000 */
    8.57980966567993164062E-01,  /* 0x3FEB749480000000 */
    8.70364665985107421875E-01,  /* 0x3FEBDA0700000000 */
    8.82643043994903564453E-01,  /* 0x3FEC3E9CA0000000 */
    8.94817709922790527344E-01,  /* 0x3FECA258C0000000 */
    9.06890571117401123047E-01,  /* 0x3FED053F60000000 */
    9.18863236904144287109E-01,  /* 0x3FED6753E0000000 */
    9.30737316608428955078E-01,  /* 0x3FEDC899A0000000 */
    9.42514479160308837891E-01,  /* 0x3FEE291420000000 */
    9.54196274280548095703E-01,  /* 0x3FEE88C6A0000000 */
    9.65784251689910888672E-01,  /* 0x3FEEE7B460000000 */
    9.77279901504516601562E-01,  /* 0x3FEF45E080000000 */
    9.88684654235839843750E-01,  /* 0x3FEFA34E00000000 */
    1.00000000000000000000E+00}; /* 0x3FF0000000000000 */

  static const double log2_tail_table[65] = {
    0.00000000000000000000E+00,  /* 0x0000000000000000 */
    3.35335459194866276130E-10,  /* 0x3DF70B48A629B89C */
    2.50974623855642191448E-09,  /* 0x3E258EFD87313606 */
    7.36565512335896390543E-09,  /* 0x3E3FA29F9AB3CF74 */
    6.23647298369351871453E-09,  /* 0x3E3AC913167CCC53 */
    1.57952684118980398844E-10,  /* 0x3DE5B5774C7658A0 */
    5.98534121117166302413E-09,  /* 0x3E39B4F3C72C4F78 */
    1.15083837244190179789E-09,  /* 0x3E13C570D0FA8F90 */
    1.20996228487473215213E-08,  /* 0x3E49FBD6879FA00B */
    7.53609938318432874467E-09,  /* 0x3E402F03B2FDD224 */
    6.38269581457264661091E-09,  /* 0x3E3B69D927DFC23C */
    5.67952040356156465017E-09,  /* 0x3E3864B2DF91E96A */
    1.16459094073677371864E-08,  /* 0x3E4902680A23A8D9 */
    2.49821633265319561946E-08,  /* 0x3E5AD30B75CB0970 */
    1.02955826924364881206E-08,  /* 0x3E461C0E73048B72 */
    1.36380709420054099385E-08,  /* 0x3E4D499BD9B32266 */
    1.09907253849796912371E-08,  /* 0x3E479A3715FC9256 */
    2.41992456974946430426E-08,  /* 0x3E59FBD6879FA00B */
    1.26785390154315961619E-08,  /* 0x3E4B3A19BCAF1AA4 */
    2.87440583546118995874E-09,  /* 0x3E28B0E2A19575B0 */
    8.49500784748945819113E-09,  /* 0x3E423E2E1169656A */
    9.77661777174938265384E-09,  /* 0x3E44FEC0A13AF880 */
    2.16260023578294509223E-08,  /* 0x3E573882A0A4146E */
    1.22576485902594488001E-08,  /* 0x3E4A52B6ACFCFDCA */
    1.85225112644675216321E-10,  /* 0x3DE975077F1F5F00 */
    1.77290105086271740075E-09,  /* 0x3E1E754D20C519E0 */
    8.18918703987935816281E-09,  /* 0x3E41960D9D9C3262 */
    2.15921288850262793860E-08,  /* 0x3E572F32FE672868 */
    1.55872185666914610882E-09,  /* 0x3E1AC754CB104AE0 */
    4.96067191344004864525E-08,  /* 0x3E6AA1E5439E15A5 */
    5.69693854190458063634E-08,  /* 0x3E6E95D00DE3B513 */
    3.75377959861950863279E-08,  /* 0x3E642727C8080ECC */
    1.35003920212974864041E-08,  /* 0x3E4CFDEB43CFD004 */
    1.99405761661543437744E-08,  /* 0x3E5569394D90D724 */
    4.95938764277745619566E-08,  /* 0x3E6AA02166CCCAB2 */
    5.69883315429349605246E-08,  /* 0x3E6E986AA9670761 */
    2.19814507699593824742E-08,  /* 0x3E579A3715FC9256 */
    1.31015261824841576777E-08,  /* 0x3E4C22A3E377A524 */
    1.22862844080671745121E-08,  /* 0x3E4A6274CF0E362C */
    3.74714569064514928410E-08,  /* 0x3E641E09AEB8CB1A */
    2.65975130588343109077E-08,  /* 0x3E5C8F11979A5DB6 */
    1.94857332324691494283E-08,  /* 0x3E54EC3293B2FBE0 */
    4.15459262300620263689E-08,  /* 0x3E664E0753C6E578 */
    4.69518366451302198484E-08,  /* 0x3E693502E409EAB7 */
    4.05011760638924658298E-08,  /* 0x3E65BE7072DBDC04 */
    2.05395606358225316367E-08,  /* 0x3E560DDF1680DD44 */
    4.09782728853196822622E-08,  /* 0x3E660002CCFE43F5 */
    3.47842490297177925737E-08,  /* 0x3E62ACB5F5EFEC49 */
    5.45992606015825934783E-08,  /* 0x3E6D5010B3666559 */
    3.64241293587091694274E-08,  /* 0x3E638E181ED0798D */
    4.00474626225128781862E-08,  /* 0x3E658014D37CDE37 */
    1.25494472416488406547E-08,  /* 0x3E4AF321AF5E9BB4 */
    2.85595789566572715872E-08,  /* 0x3E5EAA65B49696E2 */
    5.35982971014292903334E-08,  /* 0x3E6CC67E6DB516DE */
    5.36693769435427990824E-09,  /* 0x3E370D02A99B4C58 */
    5.33851529883522815863E-08,  /* 0x3E6CA9331635FEE3 */
    2.44911174062771809500E-08,  /* 0x3E5A4C112CE6312E */
    3.70450225289350432643E-10,  /* 0x3DF975077F1F5F00 */
    2.09544573213940723936E-08,  /* 0x3E567FEAD8BCCE74 */
    2.61789310367290825660E-08,  /* 0x3E5C1C02803F7554 */
    3.61063271131029934309E-08,  /* 0x3E63626A72AA21A3 */
    3.29721761549390770201E-08,  /* 0x3E61B3A9507D6DC1 */
    2.19953998687869412865E-08,  /* 0x3E579E0CAA9C9AB6 */
    3.25363260095299997864E-08,  /* 0x3E6177C23362928B */
    0.00000000000000000000E+00}; /* 0x0000000000000000 */

  static const double
    /* Reciprocal of log(2.0) */
    reclog2_lead =     1.44269504088896338700E+00, /* 0x3FF71547652B82FE */
    reclog2_tail =     2.03552737409310207851E-17; /* 0x3C7777D0FFDA0D20 */

  const double large = 1.79769313486231570815e+308; /* 0x7fefffffffffffff */

  unsigned long ux, ax, uy, ay, mask;
  int yexp, inty, xpos, ypos, negateres;
  double r, u1, u2, w, w1, w2;
  volatile int dummy;

  double u, r1, r2, f, z1, z2, q, f1, f2, poly;
  int xexp, expadjust, index, iw;

  double argx = x, argy = y;

  GET_BITS_DP64(x, ux);
  ax = ux & (~SIGNBIT_DP64);
  xpos = ax == ux;
  GET_BITS_DP64(y, uy);
  ay = uy & (~SIGNBIT_DP64);
  ypos = ay == uy;

  if (ux == 0x3ff0000000000000)
    {
      /* x = +1.0. Return +1.0 for all y, even NaN,
         raising invalid only if y is a signalling NaN */
      if (y + 1.0 == 2.0) dummy = 1;
      return 1.0;
    }
  else if (ay == 0)
    {
      /* y is zero. Return 1.0, even if x is infinity or NaN,
         raising invalid only if x is a signalling NaN */
      if (x + 1.0 == 2.0) dummy = 1;
      return 1.0;
    }
  else if (((ax & EXPBITS_DP64) == EXPBITS_DP64) &&
           (ax & MANTBITS_DP64))
    /* x is NaN. Return NaN, with invalid exception if it's
       a signalling NaN. */
    return x + x;
  else if (((ay & EXPBITS_DP64) == EXPBITS_DP64) &&
             (ay & MANTBITS_DP64))
   /* y is NaN. Return NaN, with invalid exception if y
      is a signalling NaN. */
   return y + y;
  else if (uy == 0x3ff0000000000000)
    /* y is 1.0; return x */
    return x;
  else if ((ay & EXPBITS_DP64) > 0x43e0000000000000)
    {
      /* y is infinite or so large that the result would
         overflow or underflow. Flags should be raised
         unless y is an exact infinity. */
      int yinf = (ay == EXPBITS_DP64);
      if (ypos)
        {
          /* y is +ve */
          if (ax == 0)
            /* abs(x) = 0.0. */
            return 0.0;
          else if (ax < 0x3ff0000000000000)
            {
              /* abs(x) < 1.0 */
              if (yinf)
                return 0.0;
              else
                return retval_errno_erange_underflow(argx, argy, 1);
            }
          else if (ax == 0x3ff0000000000000)
            /* abs(x) = 1.0. */
            return 1.0;
          else
            {
              /* abs(x) > 1.0 */
              if (yinf)
                return infinity_with_flags(0);
              else
                return retval_errno_erange_overflow(argx, argy, 1);
            }
        }
      else
        {
          /* y is -ve */
          if (ax == 0)
            /* abs(x) = 0.0. Return +infinity. */
            return retval_errno_edom(argx, argy, 1);
          else if (ax < 0x3ff0000000000000)
            {
            /* abs(x) < 1.0; return +infinity. */
              if (yinf)
                return infinity_with_flags(0);
              else
                return retval_errno_erange_overflow(argx, argy, 1);
            }
          else if (ax == 0x3ff0000000000000)
            /* abs(x) = 1.0. */
            return 1.0;
          else
            {
              /* abs(x) > 1.0 */
              if (yinf)
                return 0.0;
              else
                return retval_errno_erange_underflow(argx, argy, 1);
            }
        }
    }

  /* See whether y is an integer.
     inty = 0 means not an integer.
     inty = 1 means odd integer.
     inty = 2 means even integer.
  */
  yexp = ((uy & EXPBITS_DP64) >> EXPSHIFTBITS_DP64) - EXPBIAS_DP64 + 1;
  if (yexp < 1)
    inty = 0;
  else if (yexp > 53)
    inty = 2;
  else /* 1 <= yexp <= 53 */
    {
      /* Mask out the bits of r that we don't want */
      mask = (1L << (53 - yexp)) - 1;
      if ((uy & mask) != 0)
        inty = 0;
      else if (((uy & ~mask) >> (53 - yexp)) & 0x0000000000000001)
        inty = 1;
      else
        inty = 2;
    }

  if ((ax & EXPBITS_DP64) == EXPBITS_DP64)
    {
      /* x is infinity (NaN was already ruled out). */
      if (xpos)
        {
          /* x is +infinity */
          if (ypos)
            /* y > 0.0 */
            return x;
          else
            return 0.0;
        }
      else
        {
          /* x is -infinity */
          if (inty == 1)
            {
              /* y is an odd integer */
              if (ypos)
                /* Result is -infinity */
                return x;
              else
                return -zero_with_flags(0);
            }
          else
            {
              if (ypos)
                /* Result is +infinity */
                return -x;
              else
                return 0.0;
            }
        }
    }
  else if (ax == 0)
    {
      /* x is zero */
      if (xpos)
        {
          /* x is +0.0 */
          if (ypos)
            /* y is positive; return +0.0 for all cases */
            return x;
          else
            /* y is negative; return +infinity with div-by-zero
               for all cases */
            return retval_errno_edom(argx, argy, 1);
        }
      else
        {
          /* x is -0.0 */
          if (ypos)
            {
              /* y is positive */
              if (inty == 1)
                /* -0.0 raised to a positive odd integer returns -0.0 */
                return x;
              else
                /* Return +0.0 */
                return -x;
            }
          else
            {
              /* y is negative */
              if (inty == 1)
                /* -0.0 raised to a negative odd integer returns -infinity
                   with div-by-zero */
                return retval_errno_edom(argx, argy, 2);
              else
                /* Return +infinity with div-by-zero */
                return retval_errno_edom(argx, argy, 1);
            }
        }
    }

  negateres = 0;
  if (!xpos)
    {
      /* x is negative */
      if (inty)
        {
          /* It's OK because y is an integer. */
          ux = ax;
          PUT_BITS_DP64(ux, x); /* x = abs(x) */
          /* If y is odd, the result will be negative */
          negateres = (inty == 1);
        }
      else
        /* y is not an integer. Return a NaN. */
        return retval_errno_edom(argx, argy, 3);
    }

  if (ay < 0x3c00000000000000)   /* abs(y) < 2^(-63) */
      {
        /* y is close enough to zero for the result to be 1.0
           no matter what the size of x */
        return 1.0 + y;
      }

  /*
    Calculate log2(x)

    First, we decompose the argument x to the form
    x  =  2**M  *  (F1  +  F2),
    where  1 <= F1+F2 < 2, M has the value of an integer,
    F1 = 1 + j/64, j ranges from 0 to 64, and |F2| <= 1/128.

    Second, we approximate log2( 1 + F2/F1 ) by a polynomial
    in U, where U  =  2 F2 / (2 F1 + F2).
    Note that log2( 1 + F2/F1 ) = log2( 1 + U/2 ) - log2( 1 - U/2 ).
    The core approximation calculates
    Poly = [log2( 1 + U/2 ) - log2( 1 - U/2 )]/U   -   1.

    It is not hard to see that
    log2(x) = M + log2(F1) + log2( 1 + F2/F1 ).
    Hence, we return Z1 = log2(F1), and  Z2 = log2( 1 + F2/F1).
    The values of log2(F1) are calculated beforehand and stored
    in the program.
  */

  if (ux < IMPBIT_DP64)
      {
        /* The input argument x is denormalized */
        /* Normalize f by increasing the exponent by 60
           and subtracting a correction to account for the implicit
           bit. This replaces a slow denormalized
           multiplication by a fast normal subtraction. */
        static const double corr = 2.5653355008114851558350183e-290; /* 0x03d0000000000000 */
        PUT_BITS_DP64(ux | 0x03d0000000000000, f);
        f -= corr;
        GET_BITS_DP64(f, ux);
        expadjust = 60;
      }
    else
      {
        f = x;
        expadjust = 0;
      }

  /* Store the exponent of x in xexp and put
     f into the range [0.5,1) */
  xexp = (int)((ux & EXPBITS_DP64) >> EXPSHIFTBITS_DP64) - EXPBIAS_DP64 - expadjust;
  PUT_BITS_DP64((ux & MANTBITS_DP64) | HALFEXPBITS_DP64, f);

  /* Now  x = 2**xexp  * f,  1/2 <= f < 1. */

  /* Set index to be the nearest integer to 128*f */
  /*
    r = 128.0 * f;
    index = (int)(r + 0.5);
  */
  /* This code instead of the above can save several cycles.
     It only works because 64 <= r < 128, so
     the nearest integer is always contained in exactly
     7 bits, and the right shift is always the same. */
      index = (((ux & 0x000fc00000000000) | 0x0010000000000000) >> 46)
        + ((ux & 0x0000200000000000) >> 45);

  z1 = log2_lead_table[index-64];
  q = log2_tail_table[index-64];
  f1 = index * 0.0078125; /* 0.0078125 = 1/128 */
  f2 = f - f1;
  /* At this point, x = 2**xexp * ( f1  +  f2 ) where
     f1 = j/128, j = 64, 65, ..., 128 and |f2| <= 1/256. */

  /* Compute z2 from Taylor series of log2(1+f1/f2) */

  u = f2/f1;
#if 0
  div2(f2,0.0,f1,0.0,&r1,&r2);
#else
  {
    double cc, h, hh;
    mul12(u, f1, &h, &hh);
    cc = ((f2 - h) - hh) / f1;
    r1 = u + cc;
    r2 = (u - r1) + cc;
  }
#endif
  poly = -u*(u*(1./2.-u*(1./3.-u*
                         (1./4.-u*(1./5.-u*(1./6.-u*
                                            (1./7.-u*(1./8.))))))));
#if 0
  add2(r1,r2,poly,0.0,&r1,&r2);
#else
  {
    double r, s;
    r = r1 + poly;
    s = ((r1 - r) + poly) + r2;
    r1 = r + s;
    r2 = (r - r1) + s;
  }
#endif

  mul2(reclog2_lead,reclog2_tail,r1,r2,&r1,&r2);

#if 0
  add2(r1,r2,q,0.0,&r1,&r2);
#else
  {
    if (r1 == 0.0)
      r1 = q;
    else if (q != 0.0)
      {
        double r, s;
        r = r1 + q;
        s = ((r1 - r) + q) + r2;
        r1 = r + s;
        r2 = (r - r1) + s;
      }
  }
#endif

#if 0
  add2(z1,0.0,r1,r2,&z1,&z2);
#else
  {
    double r, s;
    r = z1 + r1;
    s = ((z1 - r) + r1) + r2;
    z1 = r + s;
    z2 = (r - z1) + s;
  }
#endif

  /* Now z1,z2 is an extra-precise approximation of log2(2f).
     Add xexp to z1,z2 to get the result log2(x).
     The computed r1 is not subject to rounding error because
     xexp has at most 10 significant bits, log(2) has 24 significant
     bits, and z1 has up to 24 bits; and the exponents of z1
     and z2 differ by at most 6. */

#if 0
  add2(z1,z2,xexp,0.0,&u1,&u2);
#else
  {
    double r, s;
    r = z1 + xexp;
    s = ((xexp - r) + z1) + z2;
    u1 = r + s;
    u2 = (r - u1) + s;
  }
#endif

  /* end of log2(x) calculation*/

  /* Test for overflow and underflow due to y*log2(x)
     being too large or small. */

  if ((u1+u2)*y > 1025)
    {
      if (negateres)
        return retval_errno_erange_overflow(argx, argy, -1);
      else
        return retval_errno_erange_overflow(argx, argy, 1);
    }
  else if ((u1+u2)*y < -1074)
    {
      if (negateres)
        return retval_errno_erange_underflow(argx, argy, -1);
      else
        return retval_errno_erange_underflow(argx, argy, 1);
    }

  /* Carefully compute log2(x) * y */
#if 0
  mul2(u1, u2, y, 0.0, &w1, &w2);
#else
  {
    double c, cc;
    mul12(u1, y, &c, &cc);
    cc = u2 * y + cc;
    w1 = c + cc;
    w2 = (c - w1) + cc;
  }
#endif

  w = w1 + w2;
  iw = (int)(w);

#if 0
  sub2(w1, w2, (double)iw, 0.0, &w1, &w2);
#else
  {
    double a, b;
    a = w1 - iw;
    b = ((w1 - a) - iw) + w2;
    w1 = a + b;
    w2 = (a - w1) + b;
  }
#endif

  w = w1 + w2;

  /* The following code computes r = exp2(w) */

  {
    static const double
      log2 = 6.931471805599453094178e-01, /* 0x3fe62e42fefa39ef */
      log2_lead = 6.93147167563438415527E-01, /* 0x3fe62e42f8000000 */
      log2_tail = 1.29965068938898869640E-08, /* 0x3e4be8e7bcd5e4f1 */
      one_by_32_lead = 0.03125;

    double p, z1, z2, z, hx, tx, y1, y2;
    int m, n;

    GET_BITS_DP64(w, ux);
    ax = ux & (~SIGNBIT_DP64);

    /* Handle small arguments separately */
    if (ax < 0x3fb7154764ee6c2f)   /* abs(x) < 1/(16*log2) */
      {
        if (ax < 0x3c00000000000000)   /* abs(x) < 2^(-63) */
          z = 1.0 + w; /* Raises inexact if x is non-zero */
        else
          {
            /* Split x into hx (head) and tx (tail). */
            unsigned long u;
            hx = w;
            GET_BITS_DP64(hx, u);
            u &= 0xfffffffff8000000;
            PUT_BITS_DP64(u, hx);
            tx = w - hx;
            /* Carefully multiply x by log2. y1 is the most significant
               part of the result, and y2 the least significant part */
            y1 = w * log2_lead;
            y2 = (((hx * log2_lead - y1) + hx * log2_tail) +
                  tx * log2_lead) + tx * log2_tail;

            p = y1 + y2;
            z = (9.99564649780173690e-1 +
                 (1.61251249355268050e-5 +
                  (2.37986978239838493e-2 +
                   2.68724774856111190e-7*p)*p)*p)/
              (9.99564649780173692e-1 +
               (-4.99766199765151309e-1 +
                (1.070876894098586184e-1 +
                 (-1.189773642681502232e-2 +
                  5.9480622371960190616e-4*p)*p)*p)*p);

            z = ((z * y1) + (z * y2)) + 1.0;
          }
        r = scaleDouble_2(z, iw);
      }
    else
      {
        /* Find m, z1 and z2 such that exp2(x) = 2**m * (z1 + z2) */
        splitexp(w, log2, 32.0, one_by_32_lead, 0.0, &m, &z1, &z2);

        /* Scale (z1 + z2) by 2.0**(m + iw) */

        n = m+iw;
        z = z1+z2;

        if (n < -1022)
          { /* Result will be denormalised after scaling
               down by 2**n. Using scaleDownDouble instead
               of scaleDouble_3 is faster in this case. */
            GET_BITS_DP64(z, ux);
            scaleDownDouble(ux, -n, &ux);
            PUT_BITS_DP64(ux, r);
          }
        else
          r = scaleDouble_3(z, n);
      }
  }

  /* If r overflowed or underflowed we need to deal with errno */
  if (r > large)
    {
      /* Result has overflowed. */
      if (negateres)
        return retval_errno_erange_overflow(argx, argy, -1);
      else
        return retval_errno_erange_overflow(argx, argy, 1);
    }
  else if (r == 0.0)
    {
      /* Result has underflowed. */
      if (negateres)
        return retval_errno_erange_underflow(argx, argy, -1);
      else
        return retval_errno_erange_underflow(argx, argy, 1);
    }
  else
    {
      if (negateres)
        return -r;
      else
        return r;
    }
}

weak_alias (__pow, pow)
weak_alias (__pow, __ieee754_pow)
